/*
 * gw: Internet firewall bridge with Netrek UDP support
 *
 * Original:	Kevin Smith
 * Revised:	Andy McFadden
 */
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>		/* want IPPORT_RESERVED */
#include <pwd.h>
#include <signal.h>
#include "gw.h"

char *strcasestr();
int sigintHandler();


char *prog;
FILE *logfp;
int verbose = 0;

struct gateway *gates;
struct connection *connections;
static fd_set readfd_set;
static int maxfd;
static int connect_count = 0;

static char *allowed[] = {              /* add valid initiator addresses here */
    "amdahl.uts.amdahl.com",		/* (note this is only used for TCP) */
    "sde.uts",
    "tde.uts.amdahl.com",
    "xserver.uts.amdahl.com",
    NULL };
static char *domain = "amdahl";		/* used to print warnings */


main(argc, argv)
int argc;
char **argv;
{
    FILE *fp;
    int localPort, localPort2;
    int port;
    char client[100], server[100];
    char protocol[100];
    long lastactivity = time(NULL);
    char buf[100];
    struct passwd *pwent;
    char *uid;
    time_t t;
    int c;
    extern char *optarg;
    extern int optind;
    static char *gwrc = ".gwrc";
    void(*oldhup)(), (*oldterm)();
    /*void *oldhup, *oldterm;*/


    /* parse args */
    while ((c = getopt(argc, argv, "vf:")) != EOF)
	switch (c) {
	case 'v':
	    verbose++;
	    break;
	case 'f':
	    gwrc = optarg;
	    break;
	case '?':
	default:
	    fprintf(stderr, "Internet firewall bridge v%s\n", VERSION);
	    fprintf(stderr, "usage: gw [-v] [-f .gwrc]\n\n");
	    exit(2);
	}

    /* this system call may not be portable to non-BSD; try FD_SETSIZE */
    maxfd = getdtablesize();
    if (verbose) printf("--> Maximum file descriptor is %d\n", maxfd);

    FD_ZERO(&readfd_set);
    if ((fp=fopen(gwrc, "r")) == NULL) {
	fprintf(stderr, "Unable to fopen(%s)", gwrc);
	perror(" ");
	exit(1);
    }
    while (1) {
	fgets(buf, 80, fp);
	if (feof(fp)) break;
	if (buf[0] == '#' || buf[0] == '\n') continue;
	if (!strncmp(buf, "udp", 3)) {
	    sscanf(buf, "%s %d %d %s %d",
		protocol, &localPort, &localPort2, client, &port);
	    if (strcmp(protocol, "udp")) {
		fprintf(stderr, "Protocol %s not recognized!\n", protocol);
	    } else {
		UDPnewGateway(localPort, localPort2, client, port);
	    }

	} else {
	    sscanf(buf, "%s %d %s %d", protocol, &localPort, server, &port);
	    if (strcmp(protocol, "tcp")) {
		fprintf(stderr, "Protocol %s not recognized!\n", protocol);
	    } else {
		newGateway(localPort, server, port);
	    }
	}
    }

    if ((logfp = fopen(LOGFILE, "a")) == NULL) {
#ifdef FASCIST
	perror("ERROR: Can't open log file for writing");
	exit(1);
#else
	perror("Can't open log file for writing");
#endif
    } else {
	char buf[1024];

	setvbuf(logfp, NULL, _IONBF, 0);   /* no buffering */
	if ((pwent = getpwuid(getuid())) != NULL)
	    uid = pwent->pw_name;
	else
	    uid = "???";
	t = time(0);

	fseek(logfp, (off_t) 0, 2);
	fprintf(logfp, "\n***** gw %d started by %s on %s", getpid(), uid,
		ctime(&t));
	getwd(buf);
	fseek(logfp, (off_t) 0, 2);
	fprintf(logfp, "***** gw %d config file is %s/.gwrc\n", getpid(), buf);
    }

    oldhup = SIGNAL(SIGHUP, SIG_IGN);
    oldterm = SIGNAL(SIGTERM, SIG_IGN);
    (void) SIGNAL(SIGINT, sigintHandler);

    if (verbose) printf("--> Ready\n");
    for (;;) {
	if (connect_count) {
	    /*printf("signals blocked (%d)\n", connect_count);*/
	    SIGNAL(SIGHUP, SIG_IGN);
	    SIGNAL(SIGTERM, SIG_IGN);
	} else {
	    /*printf("signals enabled (%d)\n", connect_count);*/
	    SIGNAL(SIGHUP, oldhup);
	    SIGNAL(SIGTERM, oldterm);
	}
	if (readConnections() == -1) {
	    /* One hour time out */
	    if (time(NULL) - lastactivity > 3600) {
		if (logfp != NULL) {
		    fseek(logfp, (off_t) 0, 2);
		    fprintf(logfp, "***** gw %d timed out, exiting\n",getpid());
		}
		exit(0);
	    }
	} else {
	    lastactivity=time(NULL);
	}
    }
    /*NOTREACHED*/
}

newGateway(localPort, server, port)
int localPort;
char *server;
int port;
{
    struct gateway *newguy;
    char exp_server[64];
    char **cpp;

    for (cpp = allowed; *cpp != NULL; cpp++)
        if (!strcmp(server, *cpp)) {
#ifdef FASCIST
	    fprintf(stderr,
		    "ERROR: TCP connection to machine on 'allowed' list: %s\n",
		    server);
	    exit(1);
#else
	    fprintf(stderr,
		   "WARNING: TCP connection to machine on 'allowed' list: %s\n",
		    server);
#endif
	}
    expandName(server, exp_server);		/* "foo" --> "foo.bar.com" */
    if (strcasestr(exp_server, domain) != NULL) {
#ifdef FASCIST
	fprintf(stderr,
		"ERROR: TCP connection to machine in '%s' domain: %s\n",
		domain, server);
	exit(1);
#else
	fprintf(stderr,
		"WARNING: TCP connection to machine in '%s' domain: %s\n",
		domain, server);
#endif
    }

#ifdef FASCIST
    /* this is vital to the security of the State */
    if (port < IPPORT_RESERVED) {
	fprintf(stderr, "ERROR: TCP connection to privileged port (%s:%d)\n",
		server, port);
	exit(1);
    }
#else
    /* this may be unnecessary and/or excessive */
    if (port < IPPORT_RESERVED)
	fprintf(stderr, "WARNING: TCP connection to privileged port (%s:%d)\n",
		server, port);
#endif

    newguy=(struct gateway *) malloc(sizeof(struct gateway));
    newguy->next=gates;
    newguy->protocol=PROTO_TCP;
    gates=newguy;
    newguy->listenPort=localPort;
    newguy->foreignPort=port;
    strcpy(newguy->foreignMachine,server);

    /*if (verbose)*/ printf("Opening port %d as bridge to %s %d\n",
	localPort, server, port);
    newguy->socket=listenSocket( localPort );
    if (newguy->socket == -1) {
	fprintf(stderr, "Sorry, failed to get local port %d\n", port);
	exit(1);
    }
    /*printf("Opened\n");*/
    FD_SET(newguy->socket, &readfd_set);
}

UDPnewGateway(localPort, localPort2, client, port)
int localPort, localPort2;
char *client;
int port;
{
    struct gateway *newguy, *newguy2;

    /* this is important, FASCIST or not */
    if (port < IPPORT_RESERVED) {
	fprintf(stderr,
	 "ERROR: Attempt to configure privileged UDP port (%s:%d)\n",
		client, port);
	exit(1);
    }

    newguy=(struct gateway *) malloc(sizeof(struct gateway));
    newguy->next=gates;
    newguy->protocol=PROTO_UDP;
    gates=newguy;
    newguy->listenPort=localPort;		/* get from serv on this port */
    strcpy(newguy->listenMachine,"");		/* don't know server machine */
    newguy->foreignPort=port;			/* send to cli on this port */
    strcpy(newguy->foreignMachine,client);	/* send to "client" */

    newguy2=(struct gateway *) malloc(sizeof(struct gateway));
    newguy2->next=gates;
    newguy2->protocol=PROTO_UDP2;
    gates=newguy2;
    newguy2->listenPort=localPort2;		/* get from cli on this port */
    strcpy(newguy2->listenMachine,client);	/* listen to "client" */
    newguy2->foreignPort=0;			/* don't know server port yet */
    strcpy(newguy2->foreignMachine,"");		/* or server name */

    /*if (verbose) {*/
    printf("Opening %d as UDP bridge from <server> to %s(%d)\n",
	newguy->listenPort, newguy->foreignMachine, newguy->foreignPort);
    printf("    and %d as UDP bridge from %s to <server>\n",
	newguy2->listenPort, newguy2->listenMachine);
    /*}*/

    newguy->gwPair = newguy2;		/* so we can set addr after recfrom() */
    newguy2->gwPair = newguy;		/* so we can send on his socket */
    if (UDPinit(newguy, newguy2) < 0)
	exit (-1);

    FD_SET(newguy->socket, &readfd_set);
    FD_SET(newguy2->socket, &readfd_set);
}

newConnection( readSock, writeSock )
{
    struct connection *conn;

    conn=(struct connection *) malloc(sizeof(struct connection));
    conn->next=connections;
    connections=conn;
    FD_SET(readSock, &readfd_set);
    FD_SET(writeSock, &readfd_set);
    conn->readSock=readSock;
    conn->writeSock=writeSock;
}

readConnections()
{
    struct timeval timeout;
    fd_set readfds;
    int cc;
    struct gateway *tempgate;
    struct connection *tempconn;

    timeout.tv_sec=60;	
    timeout.tv_usec=0;
    bcopy(&readfd_set, &readfds, sizeof(fd_set));
    if ((cc = select(maxfd, &readfds, NULL, NULL, &timeout)) <= 0) {
	if (cc < 0) perror("readConnections: select");
	return(-1);
    }

    for (tempconn=connections; tempconn; tempconn=tempconn->next) {
	if (tempconn->readSock == -1) continue;
	if (FD_ISSET(tempconn->readSock, &readfds)) {
	    handleUserConnData(tempconn);
	}
	if (FD_ISSET(tempconn->writeSock, &readfds)) {
	    handleServerConnData(tempconn);
	}
    }
    for (tempgate=gates; tempgate; tempgate=tempgate->next) {
	if (FD_ISSET(tempgate->socket, &readfds)) {
	    if (tempgate->protocol != PROTO_TCP)
		UDPhandleData(tempgate);
	    else
		handleNewConnReq(tempgate);
	}
    }
    return(1);
}

handleNewConnReq(gateway)
struct gateway *gateway;
{
    int readSock;
    int writeSock;
    char sorryMsg[100];
    char **cpp, namebuf[64];

    if (gateway->protocol != PROTO_TCP) {
	fprintf(stderr, "Bug!\n");
	return;
    }

    readSock=acceptConnection(gateway->socket);
    if (readSock == -1) {
	fprintf(stderr, "Bogus!  Failed to accept connection!\n");
	return;
    }

    /* check authorization */
    getPeerName(readSock, namebuf);
    for (cpp = allowed; *cpp != NULL; cpp++)
        if (!strcmp(namebuf, *cpp)) goto okay;   /* hey, I'm lazy */
    fprintf(stderr, "Connection from host %s denied\n", namebuf);
    if (logfp != NULL) {
	fseek(logfp, (off_t) 0, 2);
	fprintf(logfp, "gw %d: denied connection from host %s\n", getpid(),
	    namebuf);
    }
    close(readSock);
    return;
okay:

    writeSock=connectSocket(gateway->foreignMachine, gateway->foreignPort);
    if (writeSock == -1) {
	sprintf(sorryMsg, "Connection refused by %s %d\n\r",
	    gateway->foreignMachine, gateway->foreignPort);
	gwrite(readSock, sorryMsg, strlen(sorryMsg));
	close(readSock);
	fprintf(stderr, "Connection refused by machine %s, port %d\n",
	    gateway->foreignMachine,
	    gateway->foreignPort);
	return;
    }
    if (verbose) printf("New connection to machine %s, port %d\n",
	gateway->foreignMachine,
	gateway->foreignPort);
    newConnection(readSock, writeSock);
    connect_count++;

    if (logfp != NULL) {
	fseek(logfp, (off_t) 0, 2);
	fprintf(logfp, "gw %d(%d): TCP from %s to %s (%d)\n", getpid(),
	    readSock, namebuf, gateway->foreignMachine, gateway->foreignPort);
    }
}

handleServerConnData(conn)
struct connection *conn;
{
    if (copyData(conn->writeSock, conn->readSock)) return;
    closeConn(conn);
}

handleUserConnData(conn)
struct connection *conn;
{
    if (copyData(conn->readSock, conn->writeSock)) return;
    closeConn(conn);
}

closeConn(conn)
struct connection *conn;
{
    if (verbose) printf("Connection closed...\n");
    if (logfp != NULL) {
	fseek(logfp, (off_t) 0, 2);
	fprintf(logfp, "gw %d(%d): TCP closed\n", getpid(), conn->readSock);
    }
    FD_CLR(conn->readSock, &readfd_set);
    FD_CLR(conn->writeSock, &readfd_set);
    close(conn->readSock);
    close(conn->writeSock);
    conn->readSock = -1;
    conn->writeSock = -1;
    connect_count--;
}

copyData(readSock, writeSock)
int readSock, writeSock;
{
    char buf[16384];
    int len;

    len=read(readSock, buf, 16384);
    if (len <= 0) return(0);
    if (gwrite(writeSock, buf, len) != len) return(0);
    return(1);
}

/*
 * Find first occurrence of string.  No claims on optimality are made.
 * "pat" should have nonzero length.
 */
char *
strcasestr(str, pat)
register char *str, *pat;
{
    register char *sp, *pp;

    while (*str) {
	sp = str;
	pp = pat;
	while (toupper(*(sp++)) == toupper(*(pp++))) {
	    if (!(*sp)) return (NULL);          /* ran out of string */
	    if (!(*pp)) return (str);           /* we found a match! */
	}
	str++;
    }
    return (NULL);
}

int
sigintHandler()
{
    if (connect_count) {
	fprintf(stderr,
		"\nWARNING: gw is still in use (possibly by other users).\n");
	fprintf(stderr, "Hit ^C again to kill gw.\n");
	fprintf(stderr, "You may want to use ^Z bg and leave it running ");
	fprintf(stderr, "(gw will time out eventually)\n\n");
	SIGNAL(SIGINT, SIG_DFL);
    } else {
	exit(3);
    }
}

