/*
 * relay.h - global vars and configuration stuff
 *
 * MetaServerII relay
 */
#define DEBUG
#define DEBUG2

/* files - adjust to taste */
#define LOGFILE		"relaylog"
#define CONFIGFILE	"relayrc"

/* some time constants, in seconds */
#define TIME_RES	30		/* time between wake-up calls */
#define MAX_USER	20		/* after this long, bye bye user */

/* flood controls */
#define MAX_CALLS	3		/* you get this many calls... */
#define FLOOD_TIME	10		/* ...over this many seconds. */


/*
 * end of configurable stuff
 */

/* what version we are */
#define VERSION	"0.1"

#ifndef TRUE
# define TRUE	1
# define FALSE	0
#endif

#ifdef DEBUG
# define DBUG(x) printf x
#else
# define DBUG(x)
#endif

/* deal with funky SVR3 signal mechanism */
#ifdef _UTS
# define SYSV
#endif
#ifdef SYSV
# define SIGNAL(sig,action) sigset(sig, action)
#else
# define SIGNAL(sig,action) signal(sig, action)
#endif

/* global structure declarations */


/* port stuff */
typedef enum { DI_BUFCP } DISPLAY;
typedef struct {
    DISPLAY port_kind;
    char *port_name;
} PORT_DEF;

/* data buffer, with goodies */
typedef enum { BS_EMPTY, BS_READING, BS_VALID, BS_EXPIRED } BUF_STATE;
typedef struct {
    BUF_STATE bstatus;		/* buffer status */
    char *buf;
    int  buf_size;		/* size of the buffer */
    int  data_size;		/* size of the stuff in the buffer */
    int  ref_count;		/* number of users referencing this buf */
    time_t expire;		/* when does this data become worthless */
} DATA_BUF;

typedef enum { WD_UNKNOWN, WD_CONN, WD_READ, WD_GARBAGE, WD_TIMEOUT } WHY_DEAD;
typedef enum { CS_CLOSED, CS_CONNECTING, CS_CONNECTED } CONN_STATE;
typedef struct {
    /* static declaration stuff */
    int  port;		/* local port */
    DISPLAY kind;	/* what kind of service */
    char rhost[64];	/* remote host to contact */
    long raddr;		/* IP address for rhost (evaluated at startup) */
    int  rport;		/* remote port */
    int  stale;		/* bufcp: time until stale */
    DATA_BUF *data;	/* bufcp: current info for this port */
    /* other fields as needed for different services */

    /* connection stuff */
    CONN_STATE cstate;          /* what our (user) connection status is */
    WHY_DEAD why_dead;		/* if unable to connect, why? */
    int  sock;			/* socket #; -1 if not connected */
} PORT;

/* user stuff */
typedef struct {
    DATA_BUF *data;
    int  pos;			/* current offset into data->buf */
    time_t start_time;          /* if nonzero, then user is active */
} USER;


/* assorted library routines */
extern char *malloc(),
	    *realloc();
extern char *ctime();


/* function prototypes */
extern void init_connections(/*void*/),
	    main_loop(/*void*/);
extern int wakeup();

/* globals */
extern int verbose;
extern char *prog;
extern USER *users;
extern int port_count;
extern PORT *ports;

