/*#define UDPDEBUG*/
/*
 * New socket.c with UDP stuff
 */
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include "gw.h"

extern char *inet_ntoa();

#ifdef UDPDEBUG
# define UDPDIAG(x) { printf("UDP: "); printf x; }
#else
# define UDPDIAG(x)
#endif

int 
connectSocket(machine, port)
char *machine;
int port;
{
    struct sockaddr_in addr;
    struct hostent *hp;
    int sock;

    if ((sock=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("connectSocket: socket");
	return(-1);
    }

    if ((addr.sin_addr.s_addr = inet_addr(machine)) == -1) {
        if ((hp = gethostbyname(machine)) == NULL) {
 	    fprintf(stderr, "Who is %s?\n", machine);
	    close(sock);
 	    return(-1);
        } else {
	    addr.sin_addr.s_addr = *(long *) hp->h_addr;
        }
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);		

    if (connect(sock, &addr, sizeof(addr)) < 0) {
	perror("connectSocket: connect");
	close(sock);
	return(-1);
    }

    return(sock);
}

int 
listenSocket(port)
int port;
{
    struct sockaddr_in addr;
    int sock;
    extern int errno;
    int on=1;

    if ((sock=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("listenSocket: socket");
	return(0);
    }

    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;	/* hisaddr? */
    addr.sin_port = htons(port);
    if ( bind(sock, &addr, sizeof(addr) ) < 0) {
	perror("listenSocket: bind");
	return(-1);
    }

    if (listen(sock,5) < 0)
	perror("listenSocket: listen");

    return(sock);
}

int 
acceptConnection(sock)
int sock;
{
    struct timeval timeout;
    fd_set readfds;
    int len, cc;
    struct sockaddr_in naddr;	/* remote machine */
    int readsock;

    timeout.tv_sec=0;	
    timeout.tv_usec=0;
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    if ((cc = select(sock+1, &readfds, NULL, NULL, &timeout)) <= 0) {
	if (cc < 0) perror("acceptConnection: select");
	return(-1);
    }
    len=sizeof(naddr);
    readsock=accept(sock, &naddr, &len);
    return(readsock);
}

/* keep write()ing until it's all sent */
gwrite(fd, buf, bytes)
int fd;
register char *buf;
register int bytes;
{
    register long n;
    long orig = bytes;

    while (bytes) {
	n = write(fd, buf, bytes);
	if (n < 0) {
	    perror("gwrite");
	    return(-1);
	}
	bytes -= n;
	buf += n;
    }
    return(orig);
}

getPeerName(fd, buf)
int fd;
char *buf;
{
    struct sockaddr_in sin;
    struct hostent *hp;
    int length;

    length = sizeof (struct sockaddr_in);
    if (getpeername(fd, (struct sockaddr *) &sin, &length) < 0) {
	perror("unable to get peername");
	(void) strcpy(buf, "nowhere");
	return;
    }

    hp = gethostbyaddr((char *) &sin.sin_addr.s_addr, sizeof(long), AF_INET);
    if (hp != NULL)
	(void) strcpy(buf, hp->h_name);
    else
	(void) strcpy(buf, inet_ntoa(sin.sin_addr));
}

/* this only "expands" the name if the nameserver has it that way... oh well */
expandName(saddr, expaddr)
char *saddr, *expaddr;
{
    struct hostent *hp;

    hp = gethostbyname(saddr);
    if (hp != NULL)
	(void) strcpy(expaddr, hp->h_name);
    else {
	fprintf(stderr, "WARNING: unable to expand hostname '%s'\n", saddr);
	(void) strcpy(expaddr, saddr);
    }
}


/*
 * UDP gateway support
 *
 * Given the .gwrc line:
 *	udp 5000 5001 bezier.berkeley.edu tde.uts 5100
 *
 * (meaning "for UDP port 5000, take packets received from the server at
 * bezier and pass them to the client on tde.uts at port 5100.  For UDP
 * port 5001, take packets received from the client at tde.uts and pass
 * them to whoever sent packets to port 5000.")
 *
 * We create two sets of UDP recvfrom/sento pairs:
 *
 *	bind("bezier", 5000)
 *	recvfrom(&from_server)				bezier sends to 5000
 *	sendto("tde.uts", 5100)				we send to client
 *
 *	bind("tde.uts", 5001)
 *	recv()
 *	sendto(&from_server)				use result from above
 *
 * So gw will get the server's addr and port# from the initial server
 * transmission (so make sure the client is in "recvfrom()" mode).
 *
 * The client should bind() to port 5100, and connect() to gw's port 5001.
 * If the client tries to be intelligent and connect()s to the socket that
 * it gets from recvfrom() (i.e. 5000), it won't accomplish much of anything.
 *
 * Here's the tricky part: each has connected to a port on gw, and expects
 * to send and receive from that same port.  Thus all transmissions to
 * the client and server must be made from the same socket that we use to
 * receive data from that host (which makes good sense if you only have one
 * socket, but when you have two it can get a bit confusing).
 *
 * There were two possible implementations: have the struct gateway set to
 * send and receive from the same host, or set up the struct gateway as
 * usual and then use the other's socket when it's time to send.  I have
 * chosen the latter.
 */
#define UDPBUFSIZE	2048
char udpbuf[UDPBUFSIZE];

int 
UDPinit(gw1, gw2)
struct gateway *gw1, *gw2;
{
    struct sockaddr_in addr, caddr;
    struct hostent *hp;
    long client_addr;

    if ((client_addr = inet_addr(gw1->foreignMachine)) == -1) {
        if ((hp = gethostbyname(gw1->foreignMachine)) == NULL) {
 	    fprintf(stderr, "Who is %s?\n", gw1->foreignMachine);
 	    return(-1);
        } else {
	    client_addr = *(long *) hp->h_addr;
        }
    }
    if ((gw1->socket=socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("DGRAM socket(gw1)");
	return(-1);
    }
    if ((gw2->socket=socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("DGRAM socket(gw2)");
	return(-1);
    }

    /* gw1 listens to the server and sends to the client */
    UDPDIAG(("gw1 listening to server on port %d\n", gw1->listenPort));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(gw1->listenPort);
    if (bind(gw1->socket, &addr, sizeof(addr)) < 0) {
	perror("bind");
	fprintf(stderr, "Unable to get UDP/1 port %d\n", gw1->listenPort);
	close(gw1->socket);
	return(-1);
    }
    caddr.sin_family = AF_INET;
    caddr.sin_addr.s_addr = client_addr;
    caddr.sin_port = htons(gw1->foreignPort);
    bcopy(&caddr, (struct sockaddr_in *) gw1->addr, sizeof(caddr));

    /* gw2 listens to the client and sends to the server (addr not yet known) */
    UDPDIAG(("gw2 listening to client (0x%x) on port %d\n", client_addr,
	gw2->listenPort));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(gw2->listenPort);
    if (bind(gw2->socket, &addr, sizeof(addr)) < 0) {
	perror("bind");
	fprintf(stderr, "Unable to get UDP/2 port %d\n", gw2->listenPort);
	close(gw2->socket);
	return(-1);
    }
    ((struct sockaddr_in *) gw2->addr)->sin_port = 0;

    return(0);
}

/*
 * If the data is from #1 (server->client), send the data and update #2's
 * destination address.
 *
 * If the data is from #2 (client->server), make sure the destination address
 * is valid, and then send it along USING #1's SOCKET.
 */
UDPhandleData(gw)
struct gateway *gw;
{
    struct sockaddr_in from;
    int fromlen;
    int cc;
    struct hostent *hp;
    char *name;

    if (gw->protocol == PROTO_UDP) {
	fromlen = sizeof(from);
	if ((cc=recvfrom(gw->socket,udpbuf,UDPBUFSIZE,0,&from,&fromlen)) <= 0) {
	    perror("bad recvfrom");
	    UDPDIAG(("Result %d, error %d\n", cc, errno));
	    return;
	}
	UDPDIAG(("Recvfrom1 %d bytes\n", cc));

	gsendto(gw->gwPair->socket, udpbuf, cc, 0, gw->addr,
		sizeof(struct sockaddr_in));

	if (bcmp(&from, gw->gwPair->addr, sizeof(from))) {
	    if (verbose)
		printf("Detected new UDP connection on %d from 0x%x port %d\n",
		    gw->listenPort, from.sin_addr.s_addr, ntohs(from.sin_port));
	    hp = gethostbyaddr((char *) &from.sin_addr.s_addr, sizeof(long),
		AF_INET);
	    if (hp != NULL) name = hp->h_name;
	    else	    name = inet_ntoa(from.sin_addr);
	    if (logfp != NULL) {
		fseek(logfp, (off_t) 0, 2);
		fprintf(logfp, "gw %d(%d): UDP from %s (%d) to port %d\n",
		    getpid(), gw->socket, name, /*from.sin_addr.s_addr,*/
		    ntohs(from.sin_port), gw->listenPort);
	    }
	    UDPDIAG(("gw1's destination (the client) is 0x%x port %d\n",
		((struct sockaddr_in *) gw->addr)->sin_addr.s_addr,
		((struct sockaddr_in *) gw->addr)->sin_port));
	    UDPDIAG(("Setting gw2's destination (the server) to 0x%x port %d\n",
		from.sin_addr.s_addr, from.sin_port));
	    bcopy(&from, gw->gwPair->addr, sizeof(struct sockaddr));
	}

    } else {	/* PROTO_UDP2 */
	if (((struct sockaddr_in *) gw->addr)->sin_port == 0) {
	    fprintf(stderr, "BUG!  gw2 addr not set\n");
	    return;
	}
	/* (we know who it's from) */
	if ((cc=recv(gw->socket, udpbuf, UDPBUFSIZE, 0)) <= 0) {
	    perror("bad recv");
	    UDPDIAG(("Result %d, error %d\n", cc, errno));
	    return;
	}
	UDPDIAG(("Recv2 %d bytes\n", cc));

	gsendto(gw->gwPair->socket, udpbuf, cc, 0, gw->addr,
		sizeof(struct sockaddr_in));
    }
}

gsendto(s, msg, len, flags, to, tolen)
int s;
char *msg;
register int len, flags;
struct sockaddr_in *to;
int tolen;
{
    long orig = len;
    register int n;

#ifdef UDPDEBUG
    struct sockaddr_in addr;
    int alen;

    alen = sizeof(addr);
    if (getsockname(s, &addr, &alen) < 0) {
	perror("printUdpInfo: getsockname");
	return;
    }
    UDPDIAG(("LOCAL : addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
        addr.sin_family, ntohs(addr.sin_port)));

    UDPDIAG(("SendTo: addr=0x%x, family=%d, port=%d (%d)\n",
	to->sin_addr.s_addr, to->sin_family, ntohs(to->sin_port), tolen));
#endif

    for ( ; len ; len -= n, msg += n) {
        if ((n = sendto(s, msg, len, flags, to, tolen)) < 0) {
	    perror("gsendto");
            return(-1);
	}
    }
    return(orig);
}

