/*
 * svc_bufcp.c - service a buffered copy request
 *
 * MetaServerII relay
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "relay.h"

DATA_BUF *get_db();

/*
 * "idx" is file descriptor / user index
 * "port_idx" is the service index (i.e. index into "ports" array)
 */
int
svc_bufcp(idx, port_idx)
int idx, port_idx;
{
    DATA_BUF *dbp;

    if ((dbp = get_db(port_idx)) == NULL)	/* find or alloc a DATA_BUF */
	return (-1);

    if (dbp->bstatus == BS_EMPTY) {
	/* need to query MS-II for new info */
	dbp->buf = (char *) malloc(256);	/* foo */
	sprintf(dbp->buf, "\nTesting...\n\n");
	dbp->buf_size = 256;
	dbp->data_size = 13;
    }

    return (0);
}


/*
 * Find the DATA_BUF for the current port.  If it doesn't exist, or the
 * old one has expired, create a new one.
 *
 * If the returned buffer has bstatus==BS_EMPTY, then the caller should
 * prepare the buffer for reading.
 *
 * Expired buffers will be cast adrift.  If ref_count is zero, they will
 * be freed.
 */
DATA_BUF *
get_db(port_idx)
int port_idx;
{
    DATA_BUF *dbp;

    if ((dbp = ports[port_idx].data) != NULL) {
	switch(dbp->bstatus) {
	case BS_EMPTY:
	    /* this shouldn't happen */
	    log_msg("found BS_EMPTY in get_db");
	    DBUG(("found BS_EMPTY in get_db\n"));
	    return (dbp);
	case BS_READING:
	    /* another user just requested this; still loading info */
	    return (dbp);
	case BS_VALID:
	    /* have a valid buffer; check expire date */
	    if (dbp->expire > time(0)) {
		/* expired; mark (or free) old and create new */
		DBUG(("get_db expired buf on svc %d\n", port_idx));
		dbp->bstatus = BS_EXPIRED;
		if (!dbp->ref_count) free(dbp);
		break;	/* out of switch, to alloc */
	    }
	    return (dbp);
	case BS_EXPIRED:
	    /* this shouldn't happen */
	    log_msg("found BS_EXPIRED in get_db");
	    DBUG(("found BS_EXPIRED in get_db\n"));
	    if (!dbp->ref_count) free(dbp);
	    break;	/* out of switch, to alloc */
	default:
	    log_msg("bad dbp->bstatus in get_db");
	    DBUG(("bad state in get_db\n"));
	    break;	/* go do the alloc, but don't try to free */
	}
    }

    /* dbp was NULL or had expired */
    DBUG(("Allocating new DATA_BUF for port %d\n", port_idx));
    dbp = (DATA_BUF *) malloc(sizeof(DATA_BUF));
    ports[port_idx].data = dbp;
    dbp->bstatus = BS_EMPTY;
    dbp->buf = NULL;
    dbp->buf_size = 0;
    dbp->data_size = 0;
    dbp->ref_count = 0;
    dbp->expire = time(0) + (ports[port_idx].stale * 60);
    DBUG(("Set expire time to now + %d seconds\n", ports[port_idx].stale*60));

    return (dbp);
}


#ifdef FUBAR
#define INITIAL_SIZE	64000
#define MAX_PRINT	256
#ifdef BAD_IDEA
# define EXPAND_SIZE	4096		/* must be > MAX_PRINT */
#endif
/*
 * Print something to an expandable buffer.  Works like sprintf(), mostly.
 * Assumes that nothing we're passed will exceed MAX_PRINT bytes.
 *
 * When everybody has vsprintf, I'll do this right.
 */
void
Uprintf(idx, format, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7)
int idx;
char *format;
long arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7;
{
    USER *up;
    DATA_BUF *dbp;

    if (idx > 1024 || idx < 0) {
	fprintf(stderr, "Internal ERROR: bogus idx %d\n", idx);
	return;
    }

    up = &users[idx];
    dbp = up->data;
    if (!up->data_size) {
	if (up->buf != NULL) {
/* debug */
	    printf("HEY: Reassigning non-null buf for %d\n", idx);
/* endif */
	    log_msg("Reassigning non-null buf for %d\n", idx);
	    /* ought to free() here, but that might make us crash */
	}
	if ((up->buf = (char *) malloc(INITIAL_SIZE)) == NULL) {
	    fprintf(stderr, "malloc failure!\n");
	    log_msg("exit: malloc failure in Uprintf");
	    exit(1);
	}
	up->buf_size = INITIAL_SIZE;
    }
    if (up->buf_size - up->data_size < MAX_PRINT) {
#ifdef BAD_IDEA
	/* buffer too small, realloc */
	up->buf_size += EXPAND_SIZE;
	if ((up->buf = (char *) realloc(up->buf, up->buf_size)) == NULL) {
	    fprintf(stderr, "realloc failure!\n");
	    log_msg("exit: realloc failure in Uprintf");
	    exit(1);
	}
#else
	/* ran out of room; bummer */
	log_msg("NOTE: ran out of space in Uprintf()");
#endif
    }
    sprintf(up->buf + up->data_size, format, arg0, arg1, arg2, arg3, arg4,
	arg5, arg6, arg7);
    up->data_size += strlen(up->buf + up->data_size);
}

#endif /*FUBAR*/
